from mindsdb_sql.parser.ast import Join, Identifier, BinaryOperation, Constant, Operation, UnaryOperation
from mindsdb_sql.parser.ast.select.star import Star


def get_alias(element):
    if '.'.join(element.parts) == '*':
        return '*'
    return '.'.join(element.parts) if element.alias is None else element.alias


def identifier_to_dict(identifier):
    res = {
        'value': '.'.join(identifier.parts),
        'name': get_alias(identifier)
    }
    return res


def where_to_dict(root):
    if isinstance(root, BinaryOperation):
        op = root.op.lower()
        if op == '=':
            op = 'eq'
        return {op: [where_to_dict(root.args[0]), where_to_dict(root.args[1])]}
    elif isinstance(root, UnaryOperation):
        op = root.op.lower()
        return {op: [where_to_dict(root.args[0])]}
    elif isinstance(root, Identifier):
        return root.value
    elif isinstance(root, Constant):
        if isinstance(root.value, str):
            return {'literal': root.value}
        else:
            return root.value
    else:
        raise Exception(f'unknown type in "where": {root}')


def plain_where_conditions(self, condition):
    ''' Transform current tree view of 'where' conditions to plain dict.
        All oprations must be 'equal', and must be joined via 'AND'.
    '''
    conditions = {}
    # condition = self.mindsdb_sql_struct.where
    if condition is None:
        return conditions
    if condition.op not in ['AND', '=']:
        raise Exception('must be only AND and =')
    if isinstance(condition.args[0], Operation):
        add_condition = self._plain_where_conditions(condition.args[0])
        conditions.update(add_condition)
        add_condition = self._plain_where_conditions(condition.args[1])
        conditions.update(add_condition)
    elif isinstance(condition.args[0], Identifier):
        if isinstance(condition.args[1], Constant) is False:
            raise Exception('must be = to constant')
        conditions[condition.args[0].value] = condition.args[1].value
    return conditions


def to_moz_sql_struct(mp):
    res = {
        'select': [],
        'from': []
    }

    for t in mp.targets:
        if isinstance(t, Star):
            res['select'].append({
                'value': '*',
                'name': '*'
            })
        else:
            res['select'].append({
                'value': '.'.join(t.parts),
                'name': get_alias(t)
            })

    if isinstance(mp.from_table, Identifier):
        res['from'] = [identifier_to_dict(mp.from_table)]
    elif isinstance(mp.from_table, Join):
        if mp.from_table.join_type == 'left join':
            if not isinstance(mp.from_table.right, Identifier):
                raise Exception("only one 'level' of join supports")
            res['from'] = [
                identifier_to_dict(mp.from_table.left),
                {
                    'left join': identifier_to_dict(mp.from_table.right)
                }
            ]
            if mp.from_table.condition is not None:
                if mp.from_table.condition.op == '=':
                    res['from'][1]['on'] = {'eq': [identifier_to_dict(x) for x in mp.from_table.condition.args]}
        else:
            raise Exception('Only left join support')
    else:
        raise Exception(f'unexpected type {mp.from_table}')

    where = mp.where
    if where is not None:
        where = where_to_dict(where)
        res['where'] = where

    if mp.limit is not None:
        res['limit'] = mp.limit.value

    return res



import json
import shutil
import os

import pandas as pd

import mindsdb_datasources
from mindsdb.__about__ import __version__ as mindsdb_version
from mindsdb.interfaces.model.model_interface import ModelInterface as NativeInterface
from mindsdb_datasources import (
    FileDS, ClickhouseDS, MariaDS, MySqlDS, PostgresDS, MSSQLDS, MongoDS,
    SnowflakeDS, AthenaDS, CassandraDS, ScyllaDS
)
from mindsdb.utilities.config import Config
from mindsdb.interfaces.storage.db import session, Datasource, Semaphor, Predictor
from mindsdb.interfaces.storage.fs import FsSotre
from mindsdb.utilities.log import log
from mindsdb.interfaces.database.integrations import get_db_integration


class DataStoreWrapper(object):
    def __init__(self, data_store, company_id=None):
        self.company_id = company_id
        self.data_store = data_store

    def __getattr__(self, name):
        def wrapper(*args, **kwargs):
            if kwargs.get('company_id') is None:
                kwargs['company_id'] = self.company_id
            return getattr(self.data_store, name)(*args, **kwargs)
        return wrapper


class DataStore():
    def __init__(self):
        self.config = Config()
        self.fs_store = FsSotre()
        self.dir = self.config['paths']['datasources']
        self.mindsdb_native = NativeInterface()

    def get_analysis(self, name, company_id=None):
        datasource_record = session.query(Datasource).filter_by(company_id=company_id, name=name).first()
        if datasource_record.analysis is None:
            return None
        analysis = json.loads(datasource_record.analysis)
        return analysis

    def start_analysis(self, name, company_id=None):
        datasource_record = session.query(Datasource).filter_by(company_id=company_id, name=name).first()
        if datasource_record.analysis is not None:
            return None
        semaphor_record = session.query(Semaphor).filter_by(company_id=company_id, entity_id=datasource_record.id, entity_type='datasource').first()
        if semaphor_record is None:
            semaphor_record = Semaphor(company_id=company_id, entity_id=datasource_record.id, entity_type='datasource', action='write')
            session.add(semaphor_record)
            session.commit()
        else:
            return
        try:
            analysis = self.mindsdb_native.analyse_dataset(ds=self.get_datasource_obj(name, raw=True, company_id=company_id), company_id=company_id)
            datasource_record = session.query(Datasource).filter_by(company_id=company_id, name=name).first()
            datasource_record.analysis = json.dumps(analysis)
            session.commit()
        except Exception as e:
            log.error(e)
        finally:
            semaphor_record = session.query(Semaphor).filter_by(company_id=company_id, entity_id=datasource_record.id, entity_type='datasource').first()
            session.delete(semaphor_record)
            session.commit()

    def get_datasources(self, name=None, company_id=None):
        datasource_arr = []
        if name is not None:
            datasource_record_arr = session.query(Datasource).filter_by(company_id=company_id, name=name)
        else:
            datasource_record_arr = session.query(Datasource).filter_by(company_id=company_id)
        for datasource_record in datasource_record_arr:
            try:
                if datasource_record.data is None:
                    continue
                datasource = json.loads(datasource_record.data)
                datasource['created_at'] = datasource_record.created_at
                datasource['updated_at'] = datasource_record.updated_at
                datasource['name'] = datasource_record.name
                datasource['id'] = datasource_record.id
                datasource_arr.append(datasource)
            except Exception as e:
                log.error(e)
        return datasource_arr

    def get_data(self, name, where=None, limit=None, offset=None, company_id=None):
        offset = 0 if offset is None else offset
        ds = self.get_datasource_obj(name, company_id=company_id)

        if limit is not None:
            # @TODO Add `offset` to the `filter` method of the datasource and get rid of `offset`
            filtered_ds = ds.filter(where=where, limit=limit + offset).iloc[offset:]
        else:
            filtered_ds = ds.filter(where=where)

        filtered_ds = filtered_ds.where(pd.notnull(filtered_ds), None)
        data = filtered_ds.to_dict(orient='records')
        return {
            'data': data,
            'rowcount': len(ds),
            'columns_names': filtered_ds.columns
        }

    def get_datasource(self, name, company_id=None):
        datasource_arr = self.get_datasources(name, company_id=company_id)
        if len(datasource_arr) == 1:
            return datasource_arr[0]
        # @TODO: Remove when db swithc is more stable, this should never happen, but good santiy check while this is kinda buggy
        elif len(datasource_arr) > 1:
            log.error('Two or more datasource with the same name, (', len(datasource_arr), ') | Full list: ', datasource_arr)
            raise Exception('Two or more datasource with the same name')
        return None

    def delete_datasource(self, name, company_id=None):
        datasource_record = Datasource.query.filter_by(company_id=company_id, name=name).first()
        if not Config()["force_datasource_removing"]:
            linked_models = Predictor.query.filter_by(company_id=company_id, datasource_id=datasource_record.id).all()
            if linked_models:
                raise Exception("Can't delete {} datasource because there are next models linked to it: {}".format(name, [model.name for model in linked_models]))
        session.delete(datasource_record)
        session.commit()
        self.fs_store.delete(f'datasource_{company_id}_{datasource_record.id}')
        try:
            shutil.rmtree(os.path.join(self.dir, f'{company_id}@@@@@{name}'))
        except Exception:
            pass

    def get_vacant_name(self, base=None, company_id=None):
        ''' returns name of datasource, which starts from 'base' and ds with that name is not exists yet
        '''
        if base is None:
            base = 'datasource'
        datasources = session.query(Datasource).filter_by(company_id=company_id).all()
        datasources_names = [x.name for x in datasources]
        if base not in datasources_names:
            return base
        for i in range(1, 1000):
            candidate = f'{base}_{i}'
            if candidate not in datasources_names:
                return candidate
        raise Exception(f"Can not find appropriate name for datasource '{base}'")

    def save_datasource(self, name, source_type, source, file_path=None, company_id=None):
        if source_type == 'file' and (file_path is None):
            raise Exception('`file_path` argument required when source_type == "file"')

        datasource_record = session.query(Datasource).filter_by(company_id=company_id, name=name).first()
        while datasource_record is not None:
            raise Exception(f'Datasource with name {name} already exists')

        try:
            datasource_record = Datasource(
                company_id=company_id,
                name=name,
                datasources_version=mindsdb_datasources.__version__,
                mindsdb_version=mindsdb_version
            )
            session.add(datasource_record)
            session.commit()
            datasource_record = session.query(Datasource).filter_by(company_id=company_id, name=name).first()

            ds_meta_dir = os.path.join(self.dir, f'{company_id}@@@@@{name}')
            os.mkdir(ds_meta_dir)

            if source_type == 'file':
                source = os.path.join(ds_meta_dir, source)
                shutil.move(file_path, source)
                ds = FileDS(source)

                creation_info = {
                    'class': 'FileDS',
                    'args': [source],
                    'kwargs': {}
                }

            elif get_db_integration(source_type, company_id) is not None:
                integration = get_db_integration(source_type, company_id)

                ds_class_map = {
                    'clickhouse': ClickhouseDS,
                    'mariadb': MariaDS,
                    'mysql': MySqlDS,
                    'postgres': PostgresDS,
                    'mssql': MSSQLDS,
                    'mongodb': MongoDS,
                    'snowflake': SnowflakeDS,
                    'athena': AthenaDS,
                    'cassandra': CassandraDS,
                    'scylladb': ScyllaDS
                }

                try:
                    dsClass = ds_class_map[integration['type']]
                except KeyError:
                    raise KeyError(f"Unknown DS type: {source_type}, type is {integration['type']}")

                if dsClass is None:
                    raise Exception(f'Unsupported datasource: {source_type}, please install required dependencies!')

                if integration['type'] in ['clickhouse']:
                    creation_info = {
                        'class': dsClass.__name__,
                        'args': [],
                        'kwargs': {
                            'query': source['query'],
                            'user': integration['user'],
                            'password': integration['password'],
                            'host': integration['host'],
                            'port': integration['port']
                        }
                    }
                    ds = dsClass(**creation_info['kwargs'])

                elif integration['type'] in ['mssql', 'postgres', 'mariadb', 'mysql', 'cassandra', 'scylladb']:
                    creation_info = {
                        'class': dsClass.__name__,
                        'args': [],
                        'kwargs': {
                            'query': source['query'],
                            'user': integration['user'],
                            'password': integration['password'],
                            'host': integration['host'],
                            'port': integration['port']
                        }
                    }

                    if integration['type'] == 'mysql':
                        creation_info['kwargs']['ssl'] = integration.get('ssl')
                        creation_info['kwargs']['ssl_ca'] = integration.get('ssl_ca')
                        creation_info['kwargs']['ssl_ca_name'] = integration.get('ssl_ca_name', 'ssl_ca.pem')
                        creation_info['kwargs']['ssl_cert'] = integration.get('ssl_cert')
                        creation_info['kwargs']['ssl_cert_name'] = integration.get('ssl_cert_name', 'ssl_cert.pem')
                        creation_info['kwargs']['ssl_key'] = integration.get('ssl_key')
                        creation_info['kwargs']['ssl_key_name'] = integration.get('ssl_key_name', 'ssl_key.pem')

                    if 'database' in integration:
                        creation_info['kwargs']['database'] = integration['database']

                    if 'database' in source:
                        creation_info['kwargs']['database'] = source['database']

                    ds = dsClass(**creation_info['kwargs'])

                elif integration['type'] == 'snowflake':
                    creation_info = {
                        'class': dsClass.__name__,
                        'args': [],
                        'kwargs': {
                            'query': source['query'],
                            'schema': source['schema'],
                            'warehouse': source['warehouse'],
                            'database': source['database'],
                            'host': integration['host'],
                            'password': integration['password'],
                            'user': integration['user'],
                            'account': integration['account']
                        }
                    }

                    ds = dsClass(**creation_info['kwargs'])

                elif integration['type'] == 'mongodb':
                    if isinstance(source['find'], str):
                        source['find'] = json.loads(source['find'])
                    creation_info = {
                        'class': dsClass.__name__,
                        'args': [],
                        'kwargs': {
                            'database': source['database'],
                            'collection': source['collection'],
                            'query': source['find'],
                            'user': integration['user'],
                            'password': integration['password'],
                            'host': integration['host'],
                            'port': integration['port']
                        }
                    }

                    ds = dsClass(**creation_info['kwargs'])

                elif integration['type'] == 'athena':
                    creation_info = {
                        'class': dsClass.__name__,
                        'args': [],
                        'kwargs': {
                            'query': source['query'],
                            'staging_dir': source['staging_dir'],
                            'database': source['database'],
                            'access_key': source['access_key'],
                            'secret_key': source['secret_key'],
                            'region_name': source['region_name']
                        }
                    }

                    ds = dsClass(**creation_info['kwargs'])
            else:
                # This probably only happens for urls
                ds = FileDS(source)
                creation_info = {
                    'class': 'FileDS',
                    'args': [source],
                    'kwargs': {}
                }

            df = ds.df

            if '' in df.columns or len(df.columns) != len(set(df.columns)):
                shutil.rmtree(ds_meta_dir)
                raise Exception('Each column in datasource must have unique non-empty name')

            datasource_record.creation_info = json.dumps(creation_info)
            datasource_record.data = json.dumps({
                'source_type': source_type,
                'source': source,
                'row_count': len(df),
                'columns': [dict(name=x) for x in list(df.keys())]
            })

            self.fs_store.put(f'{company_id}@@@@@{name}', f'datasource_{company_id}_{datasource_record.id}', self.dir)
            session.commit()

        except Exception as e:
            log.error(f'Error creating datasource {name}, exception: {e}')
            try:
                self.delete_datasource(name, company_id=company_id)
            except Exception:
                pass
            raise e

        return self.get_datasource_obj(name, raw=True, company_id=company_id)

    def get_datasource_obj(self, name=None, id=None, raw=False, company_id=None):
        try:
            if name is not None:
                datasource_record = session.query(Datasource).filter_by(company_id=company_id, name=name).first()
            else:
                datasource_record = session.query(Datasource).filter_by(company_id=company_id, id=id).first()

            self.fs_store.get(f'{company_id}@@@@@{name}', f'datasource_{company_id}_{datasource_record.id}', self.dir)
            creation_info = json.loads(datasource_record.creation_info)
            if raw:
                return creation_info
            else:
                return eval(creation_info['class'])(*creation_info['args'], **creation_info['kwargs'])
        except Exception as e:
            log.error(f'Error getting datasource {name}, exception: {e}')
            return None



import os
import logging
import tempfile
from pathlib import Path

import torch.multiprocessing as mp

from mindsdb.__about__ import __version__ as mindsdb_version
from mindsdb.interfaces.database.database import DatabaseWrapper
from mindsdb.interfaces.storage.db import session, Predictor
from mindsdb.interfaces.storage.fs import FsSotre
from mindsdb.utilities.config import Config
from mindsdb.utilities.fs import create_process_mark, delete_process_mark


ctx = mp.get_context('spawn')


def create_learn_mark():
    if os.name == 'posix':
        p = Path(tempfile.gettempdir()).joinpath('mindsdb/learn_processes/')
        p.mkdir(parents=True, exist_ok=True)
        p.joinpath(f'{os.getpid()}').touch()


def delete_learn_mark():
    if os.name == 'posix':
        p = Path(tempfile.gettempdir()).joinpath('mindsdb/learn_processes/').joinpath(f'{os.getpid()}')
        if p.exists():
            p.unlink()


def run_learn(name, db_name, from_data, to_predict, kwargs, datasource_id, company_id, save):
    import mindsdb_native
    import mindsdb_datasources
    import torch
    import gc

    if 'join_learn_process' in kwargs:
        del kwargs['join_learn_process']

    create_process_mark('learn')

    config = Config()
    fs_store = FsSotre()
    mdb = mindsdb_native.Predictor(name=name, run_env={'trigger': 'mindsdb'})

    if save:
        predictor_record = Predictor.query.filter_by(company_id=company_id, name=db_name).first()
        predictor_record.datasource_id = datasource_id
        predictor_record.to_predict = to_predict
        predictor_record.native_version = mindsdb_native.__version__
        predictor_record.mindsdb_version = mindsdb_version
        predictor_record.learn_args = {
            'to_predict': to_predict,
            'kwargs': kwargs
        }
        predictor_record.data = {
            'name': db_name,
            'status': 'training'
        }
        session.commit()

    to_predict = to_predict if isinstance(to_predict, list) else [to_predict]
    data_source = getattr(mindsdb_datasources, from_data['class'])(*from_data['args'], **from_data['kwargs'])
    try:
        mdb.learn(
            from_data=data_source,
            to_predict=to_predict,
            **kwargs
        )

    except Exception as e:
        log = logging.getLogger('mindsdb.main')
        log.error(f'Predictor learn error: {e}')
        if save:
            predictor_record.data = {
                'name': db_name,
                'status': 'error'
            }
            session.commit()
        delete_process_mark('learn')

    if save:
        fs_store.put(name, f'predictor_{company_id}_{predictor_record.id}', config['paths']['predictors'])

    model_data = mindsdb_native.F.get_model_data(name)

    try:
        torch.cuda.empty_cache()
    except Exception:
        pass
    gc.collect()

    if save:
        predictor_record = Predictor.query.filter_by(company_id=company_id, name=db_name).first()
        predictor_record.data = model_data
        session.commit()

    model_data['name'] = db_name
    DatabaseWrapper(company_id).register_predictors([model_data])
    delete_process_mark('learn')


class LearnProcess(ctx.Process):
    daemon = True

    def __init__(self, *args):
        super(LearnProcess, self).__init__(args=args)

    def run(self):
        '''
        running at subprocess due to
        ValueError: signal only works in main thread

        this is work for celery worker here?
        '''
        run_learn(*self._args)



import os
from flask import request
from flask_restx import Resource, abort
from flask import current_app as ca
from mindsdb.utilities.log import log
from mindsdb.api.http.namespaces.configs.streams import ns_conf

from mindsdb.interfaces.storage.db import session
from mindsdb.interfaces.storage.db import Stream as StreamDB
from mindsdb.streams.base.base_stream import StreamTypes

from mindsdb.interfaces.database.integrations import (
    get_db_integration,
)

# def get_integration(name):
#     integrations = ca.config_obj.get('integrations', {})
#     return integrations.get(name, {})

def get_predictors():
    # full_predictors_list = [*request.native_interface.get_models(),*request.custom_models.get_models()]
    full_predictors_list = [*request.native_interface.get_models()]
    return [x["name"] for x in full_predictors_list
            if x["status"] == "complete" and x["current_phase"] == 'Trained']


def get_streams():
    streams = session.query(StreamDB).all()
    return [to_dict(stream) for stream in streams]


def to_dict(stream):
    return {"connect_info": stream.connection_params,
            "advanced_info": stream.advanced_params,
            "type": stream._type,
            "predictor": stream.predictor,
            "stream_in": stream.stream_in,
            "stream_out": stream.stream_out,
            "stream_anomaly": stream.stream_anomaly,
            "integration": stream.integration,
            "name": stream.name}


@ns_conf.route('/')
class StreamList(Resource):
    @ns_conf.doc("get_streams")
    def get(self):
        return {'streams': get_streams()}


@ns_conf.route('/<name>')
@ns_conf.param('name', 'Key-value storage stream')
class Stream(Resource):
    @ns_conf.doc("get_stream")
    def get(self, name):
        streams = get_streams()
        for stream in streams:
            if stream["name"] == name:
                return stream
        abort(404, f"Can\'t find steam: {name}")

    @ns_conf.doc("put_stream")
    def put(self, name):
        params = request.json.get('params')
        if not isinstance(params, dict):
            abort(400, "type of 'params' must be dict")
        for param in ["predictor", "stream_in", "stream_out", "integration_name"]:
            if param not in params:
                abort(400, f"'{param}' is missed.")
        integration_name = params['integration_name']
        integration_info = get_db_integration(integration_name, request.company_id, False)
        if not integration_info:
            abort(400, f"integration '{integration_name}' doesn't exist.")
        if integration_info["type"] not in ['redis', 'kafka']:
            abort(400, f"only integration of redis or kafka might be used to crate redis streams. got: '{integration_info.type}' type")
        connection_params = params.get('connect', {})
        advanced_params = params.get('advanced', {})
        predictor = params['predictor']
        stream_in = params['stream_in']
        stream_out = params['stream_out']
        stream_anomaly = params.get('stream_anomaly', stream_out)
        _type = params.get('type', 'forecast')
        if predictor not in get_predictors():
            abort(400, f"requested predictor '{predictor}' is not ready or doens't exist")
        stream = StreamDB(_type=_type, name=name, connection_params=connection_params, advanced_params=advanced_params,
                          predictor=predictor, stream_in=stream_in, stream_out=stream_out,
                          integration=integration_name, company_id=request.company_id, stream_anomaly=stream_anomaly)

        session.add(stream)
        session.commit()
        return {"status": "success", "stream_name": name}, 200

    @ns_conf.doc("delete_stream")
    def delete(self, name):
        try:
            session.query(StreamDB).filter_by(company_id=request.company_id, name=name).delete()
            session.commit()
        except Exception as e:
            log.error(e)
            abort(400, str(e))
        return '', 200



import json
import numpy as np

from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.datanode import DataNode
from mindsdb.integrations.clickhouse.clickhouse import Clickhouse
from mindsdb.integrations.postgres.postgres import PostgreSQL
from mindsdb.integrations.mariadb.mariadb import Mariadb
from mindsdb.integrations.mysql.mysql import MySQL
from mindsdb.integrations.mssql.mssql import MSSQL
from mindsdb.utilities.functions import cast_row_types
from mindsdb.utilities.config import Config
from mindsdb.interfaces.database.integrations import get_db_integration

from mindsdb.api.mysql.mysql_proxy.utilities.sql import to_moz_sql_struct, plain_where_conditions


class NumpyJSONEncoder(json.JSONEncoder):
    """
    Use this encoder to avoid
    "TypeError: Object of type float32 is not JSON serializable"

    Example:
    x = np.float32(5)
    json.dumps(x, cls=NumpyJSONEncoder)
    """
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, (np.float, np.float32, np.float64)):
            return float(obj)
        else:
            return super().default(obj)


class MindsDBDataNode(DataNode):
    type = 'mindsdb'

    def __init__(self, model_interface, custom_models, ai_table, data_store, company_id):
        self.config = Config()
        self.company_id = company_id
        self.model_interface = model_interface
        self.custom_models = custom_models
        self.ai_table = ai_table
        self.data_store = data_store

    def getTables(self):
        models = self.model_interface.get_models()
        models = [x['name'] for x in models if x['status'] == 'complete']
        models += ['predictors', 'commands']
        models += [x['name'] for x in self.custom_models.get_models()]

        ai_tables = self.ai_table.get_ai_tables()
        models += [x['name'] for x in ai_tables]
        return models

    def hasTable(self, table):
        return table in self.getTables()

    def _get_ai_table_columns(self, table_name):
        aitable_record = self.ai_table.get_ai_table(table_name)
        columns = (
            [x['name'] for x in aitable_record.query_fields] + [x['name'] for x in aitable_record.predictor_columns]
        )
        return columns

    def _get_model_columns(self, table_name):
        model = self.model_interface.get_model_data(name=table_name)
        columns = []
        columns += model['columns']
        columns += [f'{x}_original' for x in model['predict']]
        for col in model['predict']:
            if model['data_analysis_v2'][col]['typing']['data_type'] == 'Numeric':
                columns += [f"{col}_min", f"{col}_max"]
            columns += [f"{col}_confidence"]
            columns += [f"{col}_explain"]
        return columns

    def getTableColumns(self, table):
        try:
            columns = self.custom_models.get_model_data(table)['columns']
            columns += ['external_datasource', 'select_data_query', 'when_data']
            return columns
        except Exception:
            pass

        if table == 'predictors':
            return ['name', 'status', 'accuracy', 'predict', 'select_data_query', 'external_datasource', 'training_options']
        if table == 'commands':
            return ['command']

        columns = []

        ai_tables = self.ai_table.get_ai_table(table)
        if ai_tables is not None:
            columns = self._get_ai_table_columns(table)
        elif table in [x['name'] for x in self.model_interface.get_models()]:
            columns = self._get_model_columns(table)
            columns += ['when_data', 'select_data_query', 'external_datasource']

        return columns

    def _select_predictors(self):
        models = self.model_interface.get_models()
        # TODO add custom models
        return [{
            'name': x['name'],
            'status': x['status'],
            'accuracy': str(x['accuracy']) if x['accuracy'] is not None else None,
            'predict': ', '.join(x['predict']),
            'select_data_query': '',
            'external_datasource': '',  # TODO
            'training_options': ''  # TODO ?
        } for x in models]

    def delete_predictor(self, name):
        self.model_interface.delete_model(name)

    def _select_from_ai_table(self, table, columns, where):
        aitable_record = self.ai_table.get_ai_table(table)
        integration = aitable_record.integration_name
        query = aitable_record.integration_query
        predictor_name = aitable_record.predictor_name

        ds_name = self.data_store.get_vacant_name('temp')
        self.data_store.save_datasource(ds_name, integration, {'query': query})
        dso = self.data_store.get_datasource_obj(ds_name, raw=True)
        res = self.model_interface.predict(predictor_name, 'dict', when_data=dso)
        self.data_store.delete_datasource(ds_name)

        keys_map = {}
        for f in aitable_record.predictor_columns:
            keys_map[f['value']] = f['name']
        for f in aitable_record.query_fields:
            keys_map[f['name']] = f['name']
        keys = list(keys_map.keys())

        data = []
        for i, el in enumerate(res):
            data.append({keys_map[key]: el[key] for key in keys})

        return data

    def select_query(self, query):
        from mindsdb.api.mysql.mysql_proxy.utilities.sql import to_moz_sql_struct, plain_where_conditions
        moz_struct = to_moz_sql_struct(query)
        data = self.select(
            table=query.from_table.parts[-1],
            columns=None,
            where=moz_struct.get('where')
        )
        return data

    def select(self, table, columns=None, where=None, where_data=None, order_by=None, group_by=None, came_from=None, is_timeseries=False):
        ''' NOTE WHERE statements can be just $eq joined with 'and'
        '''
        if table == 'predictors':
            return self._select_predictors()
        if table == 'commands':
            return []
        if self.ai_table.get_ai_table(table):
            return self._select_from_ai_table(table, columns, where)

        # try:
        #     where = plain_where_conditions(where)
        # except Exception as e:
        #     x = 1

        original_when_data = None
        if 'when_data' in where:
            if len(where) > 1:
                raise ValueError("Should not be used any other keys in 'where', if 'when_data' used")
            try:
                original_when_data = where['when_data']['$eq']
                where_data = json.loads(where['when_data']['$eq'])
                if isinstance(where_data, list) is False:
                    where_data = [where_data]
            except Exception:
                raise ValueError(f'''Error while parse 'when_data'="{where_data}"''')
        external_datasource = None
        if 'external_datasource' in where:
            external_datasource = where['external_datasource']['$eq']
            del where['external_datasource']

        select_data_query = None
        if came_from is not None and 'select_data_query' in where:
            select_data_query = where['select_data_query']['$eq']
            del where['select_data_query']

            integration_data = get_db_integration(came_from, self.company_id)
            dbtype = integration_data['type']
            if dbtype == 'clickhouse':
                ch = Clickhouse(self.config, came_from, integration_data)
                res = ch._query(select_data_query.strip(' ;\n') + ' FORMAT JSON')
                data = res.json()['data']
            elif dbtype == 'mariadb':
                maria = Mariadb(self.config, came_from, integration_data)
                data = maria._query(select_data_query)
            elif dbtype == 'mysql':
                mysql = MySQL(self.config, came_from, integration_data)
                data = mysql._query(select_data_query)
            elif dbtype == 'postgres':
                mysql = PostgreSQL(self.config, came_from, integration_data)
                data = mysql._query(select_data_query)
            elif dbtype == 'mssql':
                mssql = MSSQL(self.config, came_from, integration_data)
                data = mssql._query(select_data_query, fetch=True)
            else:
                raise Exception(f'Unknown database type: {dbtype}')

            if where_data is None:
                where_data = data
            else:
                where_data += data

        new_where = {}
        if where_data is None:
            for key, value in where.items():
                if isinstance(value, dict) is False or len(value.keys()) != 1 or list(value.keys())[0] != '$eq':
                    # TODO value should be just string or number
                    raise Exception()
                new_where[key] = value['$eq']

            if len(new_where) == 0:
                return []

            where_data = [new_where]

        try:
            model = self.custom_models.get_model_data(name=table)
        except Exception:
            model = self.model_interface.get_model_data(name=table)

        predicted_columns = model['predict']

        original_target_values = {}
        for col in predicted_columns:
            if where_data is not None:
                if col in where_data:
                    original_target_values[col + '_original'] = list(where_data[col])
                else:
                    original_target_values[col + '_original'] = [None] * len(where_data)
            else:
                original_target_values[col + '_original'] = [None]

        if table in [x['name'] for x in self.custom_models.get_models()]:
            res = self.custom_models.predict(name=table, when_data=where_data)

            data = []
            fields = model['columns']
            for i, ele in enumerate(res):
                row = {}
                row['select_data_query'] = select_data_query
                row['external_datasource'] = external_datasource
                row['when_data'] = original_when_data

                for key in ele:
                    row[key] = ele[key]['predicted_value']
                    # FIXME prefer get int from model_interface in this case
                    if model['data_analysis_v2'][key]['typing']['data_subtype'] == 'Int':
                        row[key] = int(row[key])

                for k in fields:
                    if k not in ele:
                        if isinstance(where_data, list):
                            if k in where_data[i]:
                                row[k] = where_data[i][k]
                            else:
                                row[k] = None
                        elif k in where_data.columns:
                            row[k] = where_data[k].iloc[i]
                        else:
                            row[k] = None

                for k in original_target_values:
                    row[k] = original_target_values[k][i]

                data.append(row)

            field_types = {f: model['data_analysis_v2'][f]['typing']['data_subtype'] for f in fields if 'typing' in model['data_analysis_v2'][f]}
            for row in data:
                cast_row_types(row, field_types)

            return data
        else:
            pred_dicts, explanations = self.model_interface.predict(table, 'dict&explain', when_data=where_data)
            # Fix since for some databases we *MUST* return the same value for the columns originally specified in the `WHERE`
            if not is_timeseries:
                if isinstance(where_data, list):
                    for i in range(len(pred_dicts)):
                        for col in where_data[i]:
                            if col not in predicted_columns:
                                pred_dicts[i][col] = where_data[i][col]

                if isinstance(where_data, dict):
                    for col in where_data:
                        if col not in predicted_columns:
                            pred_dicts[0][col] = where_data[col]
            else:
                pred_dict = pred_dicts[0]
                new_pred_dicts = []
                predict = model['predict'][0]
                data_column = model['timeseries']['user_settings']['order_by'][0]
                predictions = pred_dict[predict]
                if isinstance(predictions, list) is False:
                    predictions = [predictions]
                data_values = pred_dict[data_column]
                if isinstance(data_values, list) is False:
                    data_values = [data_values]
                for i in range(model['timeseries']['user_settings']['nr_predictions']):
                    nd = {}
                    nd.update(pred_dict)
                    new_pred_dicts.append(nd)
                    nd[predict] = predictions[i]
                    nd[data_column] = data_values[-1 * (i + 1)]
                pred_dicts = new_pred_dicts

            if columns is None:
                columns = list(pred_dicts[0].keys())

            keys = [x for x in pred_dicts[0] if x in columns]
            min_max_keys = []
            for col in predicted_columns:
                if model['data_analysis_v2'][col]['typing']['data_type'] == 'Numeric':
                    min_max_keys.append(col)

            data = []
            explains = []
            for i, el in enumerate(pred_dicts):
                data.append({key: el[key] for key in keys})
                if not is_timeseries:
                    explains.append(explanations[i])
                else:
                    explains.append(explanations[0])

            field_types = {}
            for f in set(model['columns']) & set(model['data_analysis_v2'].keys()):
                if 'typing' in model['data_analysis_v2'][f]:
                    field_types[f] = model['data_analysis_v2'][f]['typing']['data_subtype']

            for i, row in enumerate(data):
                cast_row_types(row, field_types)

                row['select_data_query'] = select_data_query
                row['external_datasource'] = external_datasource
                row['when_data'] = original_when_data

                for k in original_target_values:
                    row[k] = original_target_values[k][i]

                explanation = explains[i]
                for key in predicted_columns:
                    row[key + '_confidence'] = explanation[key]['confidence']
                    row[key + '_explain'] = json.dumps(explanation[key], cls=NumpyJSONEncoder, ensure_ascii=False)
                for key in min_max_keys:
                    row[key + '_min'] = min(explanation[key]['confidence_interval'])
                    row[key + '_max'] = max(explanation[key]['confidence_interval'])

            return data



import atexit
import traceback
import sys
import os
import time
import asyncio
import signal

import torch.multiprocessing as mp

from mindsdb.utilities.config import Config, STOP_THREADS_EVENT
from mindsdb.interfaces.model.model_interface import ray_based, ModelInterface
from mindsdb.api.http.start import start as start_http
from mindsdb.api.mysql.start import start as start_mysql
from mindsdb.api.mongo.start import start as start_mongo
from mindsdb.utilities.ps import is_pid_listen_port, get_child_pids
from mindsdb.utilities.functions import args_parse
from mindsdb.interfaces.database.database import DatabaseWrapper
from mindsdb.utilities.log import log

from mindsdb.interfaces.database.integrations import get_db_integrations

COMPANY_ID = os.environ.get('MINDSDB_COMPANY_ID', None)


def close_api_gracefully(apis):
    try:
        for api in apis.values():
            process = api['process']
            childs = get_child_pids(process.pid)
            for p in childs:
                try:
                    os.kill(p, signal.SIGTERM)
                except Exception:
                    p.kill()
            sys.stdout.flush()
            process.terminate()
            process.join()
            sys.stdout.flush()
        if ray_based:
            os.system('ray stop --force')
    except KeyboardInterrupt:
        sys.exit(0)


if __name__ == '__main__':
    mp.freeze_support()
    args = args_parse()
    config = Config()

    if args.verbose is True:
        # Figure this one out later
        pass

    os.environ['DEFAULT_LOG_LEVEL'] = config['log']['level']['console']
    os.environ['LIGHTWOOD_LOG_LEVEL'] = config['log']['level']['console']

    # Switch to this once the native interface has it's own thread :/
    ctx = mp.get_context('spawn')

    from mindsdb.__about__ import __version__ as mindsdb_version
    print(f'Version {mindsdb_version}')

    print(f'Configuration file:\n   {config.config_path}')
    print(f"Storage path:\n   {config['paths']['root']}")

    # @TODO Backwards compatibiltiy for tests, remove later
    from mindsdb.interfaces.database.integrations import add_db_integration, get_db_integration, remove_db_integration
    dbw = DatabaseWrapper(COMPANY_ID)
    model_interface = ModelInterface()
    raw_model_data_arr = model_interface.get_models()
    model_data_arr = []
    for model in raw_model_data_arr:
        if model['status'] == 'complete':
            x = model_interface.get_model_data(model['name'])
            try:
                model_data_arr.append(model_interface.get_model_data(model['name']))
            except Exception:
                pass

    is_cloud = config.get('cloud', False)
    if not is_cloud:
        for integration_name in get_db_integrations(COMPANY_ID, sensitive_info=True):
            print(f"Setting up integration: {integration_name}")
            if get_db_integration(integration_name, COMPANY_ID).get('publish', False):
                # do setup and register only if it is 'publish' integration
                dbw.setup_integration(integration_name)
                dbw.register_predictors(model_data_arr, integration_name=integration_name)

    for integration_name in config.get('integrations', {}):
        try:
            it = get_db_integration(integration_name, None)
            if it is not None:
                remove_db_integration(integration_name, None)
            print(f'Adding: {integration_name}')
            add_db_integration(integration_name, config['integrations'][integration_name], None)            # Setup for user `None`, since we don't need this for cloud
            if config['integrations'][integration_name].get('publish', False) and not is_cloud:
                dbw.setup_integration(integration_name)
                dbw.register_predictors(model_data_arr, integration_name=integration_name)
        except Exception as e:
            log.error(f'\n\nError: {e} adding database integration {integration_name}\n\n')

    del model_interface
    del dbw
    # @TODO Backwards compatibiltiy for tests, remove later

    if args.api is None:
        api_arr = ['http', 'mysql']
    else:
        api_arr = args.api.split(',')

    apis = {
        api: {
            'port': config['api'][api]['port'],
            'process': None,
            'started': False
        } for api in api_arr
    }

    start_functions = {
        'http': start_http,
        'mysql': start_mysql,
        'mongodb': start_mongo
    }

    for api_name, api_data in apis.items():
        if api_data['started']:
            continue
        print(f'{api_name} API: starting...')
        try:
            if api_name == 'http':
                p = ctx.Process(target=start_functions[api_name], args=(args.verbose, args.no_studio))
            else:
                p = ctx.Process(target=start_functions[api_name], args=(args.verbose,))
            p.start()
            api_data['process'] = p
        except Exception as e:
            log.error(f'Failed to start {api_name} API with exception {e}\n{traceback.format_exc()}')
            close_api_gracefully(apis)
            raise e

    atexit.register(close_api_gracefully, apis=apis)

    async def wait_api_start(api_name, pid, port):
        timeout = 60
        start_time = time.time()
        started = is_pid_listen_port(pid, port)
        while (time.time() - start_time) < timeout and started is False:
            await asyncio.sleep(0.5)
            started = is_pid_listen_port(pid, port)
        return api_name, port, started

    async def wait_apis_start():
        futures = [
            wait_api_start(api_name, api_data['process'].pid, api_data['port'])
            for api_name, api_data in apis.items() if 'port' in api_data
        ]
        for i, future in enumerate(asyncio.as_completed(futures)):
            api_name, port, started = await future
            if started:
                print(f"{api_name} API: started on {port}")
            else:
                log.error(f"ERROR: {api_name} API cant start on {port}")

    ioloop = asyncio.get_event_loop()
    ioloop.run_until_complete(wait_apis_start())
    ioloop.close()

    try:
        for api_data in apis.values():
            api_data['process'].join()
    except KeyboardInterrupt:
        print('Stopping stream integrations...')
        STOP_THREADS_EVENT.set()
        print('Closing app...')



from mindsdb.interfaces.storage.db import session, AITable
from mindsdb.utilities.log import log


class AITableStore():
    def __init__(self, company_id=None):
        self.company_id = company_id

    def is_ai_table(self, name):
        record = self.get_ai_table(name.lower())
        return record is not None

    def get_ai_table(self, name):
        ''' get particular ai table
        '''
        aitable_record = session.query(AITable).filter_by(company_id=self.company_id, name=name.lower()).first()
        return aitable_record

    def get_ai_tables(self):
        ''' get list of ai tables
        '''
        aitable_records = [x.__dict__ for x in session.query(AITable).filter_by(company_id=self.company_id)]
        return aitable_records

    def add(self, name, integration_name, integration_query, query_fields, predictor_name, predictor_fields):
        ai_table_record = AITable(
            name=name.lower(),
            integration_name=integration_name,
            integration_query=integration_query,
            query_fields=query_fields,
            predictor_name=predictor_name,
            predictor_columns=predictor_fields,
            company_id=self.company_id
        )
        session.add(ai_table_record)
        session.commit()

    def query(self, name, where=None):
        ''' query to ai table
        '''
        pass



import json
from threading import Thread, Event
import walrus

from mindsdb.utilities.log import log
from mindsdb.utilities.cache import Cache
from mindsdb.streams.base.base_stream import StreamTypes, BaseStream
from mindsdb.interfaces.storage.db import session
from mindsdb.interfaces.storage.db import Predictor as DBPredictor
from mindsdb.interfaces.model.model_interface import ModelInterface as NativeInterface


class RedisStream(Thread, BaseStream):
    def __init__(self, name, connection_info, advanced_info, stream_in, stream_out, stream_anomaly, predictor, _type):
        self.stream_name = name
        self.connection_info = connection_info
        self.connection_info.update(advanced_info)
        self.predictor = predictor
        self.client = self._get_client()
        self.stream_in_name = stream_in
        self.stream_out_name = stream_out
        self.stream_anomaly_name = stream_anomaly
        self.stream_in = self.client.Stream(stream_in)
        self.stream_out = self.client.Stream(stream_out)
        self.stream_anomaly = self.client.Stream(stream_anomaly)
        self._type = _type

        BaseStream.__init__(self)
        if self._type.lower() == StreamTypes.timeseries:
            super().__init__(target=RedisStream.make_timeseries_predictions, args=(self,))
        else:
            super().__init__(target=RedisStream.make_predictions, args=(self,))
            self.cache = None

    def _get_client(self):
        return walrus.Database(**self.connection_info)

    def predict_ts(self, group_by):
        with self.cache:
            when_list = self.cache[group_by]
            for x in when_list:
                if self.target not in x:
                    x['make_predictions'] = False
                else:
                    x['make_predictions'] = True
            result = self.native_interface.predict(self.predictor, self.format_flag, when_data=when_list)
            log.debug(f"TIMESERIES STREAM {self.stream_name}: got {result}")
            for res in result:
                in_json = json.dumps(res)
                stream_out = self.stream_anomaly if self.is_anomaly(res) else self.stream_out
                stream_out.add({'prediction': in_json})

            #delete the oldest record from cache
            updated_list = self.cache[group_by][1:]
            self.cache[group_by] = updated_list

    def predict(self, stream_in):
        predict_info = stream_in.read()
        for record in predict_info:
            record_id = record[0]
            raw_when_data = record[1]
            when_data = self.decode(raw_when_data)
            result = self.native_interface.predict(self.predictor, self.format_flag, when_data=when_data)
            log.debug(f"STREAM {self.stream_name}: got {result}")
            for res in result:
                in_json = json.dumps(res)
                stream_out = self.stream_anomaly if self.is_anomaly(res) else self.stream_out
                stream_out.add({'prediction': in_json})
            stream_in.delete(record_id)

    def make_predictions(self):
        predict_record = session.query(DBPredictor).filter_by(company_id=self.company_id, name=self.predictor).first()
        if predict_record is None:
            log.error(f"STREAM {self.stream_name} got error: requested predictor {self.predictor} is not exist")
            return

        while not self.stop_event.wait(0.5):
            self.predict(self.stream_in)

        session.close()
        log.debug(f"STREAM {self.stream_name}: stopping...")

    def make_timeseries_predictions(self):
        self.cache = Cache(self.stream_name)
        predict_record = session.query(DBPredictor).filter_by(company_id=self.company_id, name=self.predictor).first()
        if predict_record is None:
            log.error(f"STREAM {self.stream_name} got error: requested predictor {self.predictor} is not exist")
            return
        ts_settings = self.get_ts_settings(predict_record)
        log.debug(f"STREAM {self.stream_name} TS_SETTINGS: {ts_settings}")
        self.target = ts_settings['to_predict'][0]
        self.window = ts_settings['window']
        self.gb = None if not ts_settings.get('group_by') else ts_settings.get('group_by')[0]
        self.dt = ts_settings['order_by'][0]

        while not self.stop_event.wait(0.5):
            predict_info = self.stream_in.read()
            for record in predict_info:
                record_id = record[0]
                raw_when_data = record[1]
                when_data = self.decode(raw_when_data)
                log.debug(f"STREAM {self.stream_name}: next record have read from {self.stream_in.key}: {when_data}")
                self.to_cache(when_data)
                self.stream_in.delete(record_id)
        self.cache.delete()
        session.close()

    def decode(self, redis_data):
        decoded = {}
        for k in redis_data:
            decoded[k.decode('utf8')] = redis_data[k].decode('utf8')
        return decoded



import json
from threading import Thread

import kafka

from mindsdb.utilities.log import log
from mindsdb.utilities.cache import Cache
from mindsdb.streams.base.base_stream import StreamTypes, BaseStream
from mindsdb.interfaces.storage.db import session
from mindsdb.interfaces.storage.db import Predictor as DBPredictor


class KafkaStream(Thread, BaseStream):
    def __init__(self, name, connection_info, advanced_info, topic_in, topic_out, topic_anomaly, predictor, _type):
        self.stream_name = name
        self.connection_info = connection_info
        self.advanced_info = advanced_info
        self.predictor = predictor
        self.stream_in_name = topic_in
        self.stream_out_name = topic_out
        self.stream_anomaly_name = topic_anomaly
        self.consumer = kafka.KafkaConsumer(**self.connection_info, **self.advanced_info.get('consumer', {}))
        self.consumer.subscribe(topics=[self.stream_in_name])
        self.producer = kafka.KafkaProducer(**self.connection_info, **self.advanced_info.get('producer', {}), acks='all')

        BaseStream.__init__(self)
        self._type = _type

        if self._type.lower() == StreamTypes.timeseries:
            super().__init__(target=KafkaStream.make_timeseries_predictions, args=(self,))
        else:
            super().__init__(target=KafkaStream.make_prediction, args=(self,))
            self.cache = None

    def predict_ts(self, group_by):
        with self.cache:
            when_list = self.cache[group_by]
            for x in when_list:
                if self.target not in x:
                    x['make_predictions'] = False
                else:
                    x['make_predictions'] = True
            result = self.native_interface.predict(self.predictor, self.format_flag, when_data=when_list)
            log.debug(f"TIMESERIES STREAM {self.stream_name}: got {result}")
            for res in result:
                in_json = json.dumps(res)
                to_send = in_json.encode('utf-8')
                out_stream = self.stream_anomaly_name if self.is_anomaly(res) else self.stream_out_name
                log.debug(f"STREAM {self.stream_name}: sending {to_send}")
                self.producer.send(out_stream, to_send)

            #delete the oldest record from cache
            updated_list = self.cache[group_by][1:]
            self.cache[group_by] = updated_list

    def make_timeseries_predictions(self):
        self.cache = Cache(self.stream_name)
        log.debug("STREAM: running 'make_timeseries_predictions'")
        predict_record = session.query(DBPredictor).filter_by(company_id=self.company_id, name=self.predictor).first()
        if predict_record is None:
            log.error(f"STREAM {self.stream_name} got error: requested predictor {self.predictor} is not exist")
            return
        ts_settings = self.get_ts_settings(predict_record)
        log.debug(f"STREAM {self.stream_name} TS_SETTINGS: {ts_settings}")
        self.target = ts_settings['to_predict'][0]
        self.window = ts_settings['window']
        self.gb = None if not ts_settings.get('group_by') else ts_settings.get('group_by')[0]
        self.dt = ts_settings['order_by'][0]

        while not self.stop_event.wait(0.5):
            try:
                msg_str = next(self.consumer)
                when_data = json.loads(msg_str.value)
                self.to_cache(when_data)
            except StopIteration:
                pass

        log.debug(f"STREAM {self.stream_name}: stopping...")
        self.producer.close()
        self.consumer.close()
        self.cache.delete()
        session.close()

    def make_prediction(self):
        predict_record = session.query(DBPredictor).filter_by(company_id=self.company_id, name=self.predictor).first()
        if predict_record is None:
            log.error(f"STREAM {self.stream_name} got error: requested predictor {self.predictor} is not exist")
            return
        while not self.stop_event.wait(0.5):
            try:
                msg_str = next(self.consumer)
                when_data = json.loads(msg_str.value)
                result = self.native_interface.predict(self.predictor, self.format_flag, when_data=when_data)
                log.debug(f"STREAM {self.stream_name}: got {result}")
                for res in result:
                    in_json = json.dumps(res)
                    to_send = in_json.encode('utf-8')
                    log.debug(f"STREAM {self.stream_name}: sending {to_send}")
                    out_stream = self.stream_anomaly_name if self.is_anomaly(res) else self.stream_out_name
                    self.producer.send(out_stream, to_send)
            except StopIteration:
                pass
        log.debug(f"STREAM {self.stream_name}: stopping...")
        self.producer.close()
        self.consumer.close()
        session.close()



from dateutil.parser import parse as parse_datetime
import pickle
from pathlib import Path
import psutil
import datetime
import time
import os
from contextlib import contextmanager

import pandas as pd
import mindsdb_datasources

from mindsdb.utilities.fs import create_directory, create_process_mark, delete_process_mark
from mindsdb.interfaces.database.database import DatabaseWrapper
from mindsdb.utilities.config import Config
from mindsdb.interfaces.storage.fs import FsSotre
from mindsdb.utilities.log import log
import pyarrow as pa
import pyarrow.flight as fl


class ModelController():
    def __init__(self, ray_based):
        self.config = Config()
        self.fs_store = FsSotre()
        self.predictor_cache = {}
        self.ray_based = ray_based

    def _invalidate_cached_predictors(self):
        from mindsdb_datasources import (FileDS, ClickhouseDS, MariaDS, MySqlDS,
                                         PostgresDS, MSSQLDS, MongoDS,
                                         SnowflakeDS, AthenaDS)
        from mindsdb.interfaces.storage.db import session, Predictor

        # @TODO: Cache will become stale if the respective NativeInterface is not invoked yet a bunch of predictors remained cached, no matter where we invoke it. In practice shouldn't be a big issue though
        for predictor_name in list(self.predictor_cache.keys()):
            if (datetime.datetime.now() - self.predictor_cache[predictor_name]['created']).total_seconds() > 1200:
                del self.predictor_cache[predictor_name]

    def _lock_predictor(self, id, mode='write', company_id=None):
        from mindsdb.interfaces.storage.db import session, Semaphor

        while True:
            semaphor_record = session.query(Semaphor).filter_by(company_id=company_id, entity_id=id, entity_type='predictor').first()
            if semaphor_record is not None:
                if mode == 'read' and semaphor_record.action == 'read':
                    return True
            try:
                semaphor_record = Semaphor(company_id=company_id, entity_id=id, entity_type='predictor', action=mode)
                session.add(semaphor_record)
                session.commit()
                return True
            except Exception:
                pass
            time.sleep(1)

    def _unlock_predictor(self, id, company_id=None):
        from mindsdb.interfaces.storage.db import session, Semaphor
        semaphor_record = session.query(Semaphor).filter_by(company_id=company_id, entity_id=id, entity_type='predictor').first()
        if semaphor_record is not None:
            session.delete(semaphor_record)
            session.commit()

    @contextmanager
    def _lock_context(self, id, mode='write'):
        try:
            self._lock_predictor(id, mode)
            yield True
        finally:
            self._unlock_predictor(id)

    def _setup_for_creation(self, name, original_name, company_id=None):
        from mindsdb.interfaces.storage.db import session, Predictor

        if name in self.predictor_cache:
            del self.predictor_cache[name]
        # Here for no particular reason, because we want to run this sometimes but not too often
        self._invalidate_cached_predictors()

        predictor_record = Predictor.query.filter_by(company_id=company_id, name=original_name).first()
        if predictor_record is not None:
            raise Exception(f'Predictor with name {original_name} already exists.')

        predictor_dir = Path(self.config['paths']['predictors']).joinpath(name)
        create_directory(predictor_dir)
        predictor_record = Predictor(company_id=company_id, name=original_name, is_custom=False)

        session.add(predictor_record)
        session.commit()

    def _try_outdate_db_status(self, predictor_record):
        from mindsdb import __version__ as mindsdb_version
        from mindsdb.interfaces.storage.db import session
        from packaging import version

        if predictor_record.update_status == 'update_failed':
            return predictor_record

        try:
            if version.parse(predictor_record.mindsdb_version) < version.parse(mindsdb_version):
                predictor_record.update_status = 'available'
        except Exception:
            # predictor.mindsdb_version can be None at begining of training
            pass

        session.commit()
        return predictor_record

    def create(self, name, company_id=None):
        import mindsdb_native

        original_name = name
        name = f'{company_id}@@@@@{name}'

        self._setup_for_creation(name, original_name, company_id=company_id)
        predictor = mindsdb_native.Predictor(name=name, run_env={'trigger': 'mindsdb'})
        return predictor

    def learn_for_update(self, name, from_data, to_predict, datasource_id, kwargs={}, company_id=None):
        kwargs['join_learn_process'] = True
        return self.learn(name, from_data, to_predict, datasource_id, kwargs, company_id, False)

    def learn(self, name, from_data, to_predict, datasource_id, kwargs={}, company_id=None, save=True):
        from mindsdb.interfaces.model.learn_process import LearnProcess, run_learn

        create_process_mark('learn')
        original_name = name
        name = f'{company_id}@@@@@{name}'
        join_learn_process = kwargs.get('join_learn_process', False)

        if save:
            self._setup_for_creation(name, original_name, company_id=company_id)

        if self.ray_based:
            run_learn(
                name=name,
                db_name=original_name,
                from_data=from_data,
                to_predict=to_predict,
                kwargs=kwargs,
                datasource_id=datasource_id,
                company_id=company_id,
                save=save
            )

        else:
            p = LearnProcess(name, original_name, from_data, to_predict, kwargs, datasource_id, company_id, save)
            p.start()
            if join_learn_process is True:
                p.join()
                if p.exitcode != 0:
                    delete_process_mark('learn')
                    raise Exception('Learning process failed !')

        delete_process_mark('learn')
        return 0

    def predict(self, name, pred_format, when_data=None, kwargs={}, company_id=None):
        from mindsdb_datasources import (FileDS, ClickhouseDS, MariaDS,
                                         MySqlDS, PostgresDS, MSSQLDS, MongoDS,
                                         SnowflakeDS, AthenaDS)
        import mindsdb_native
        from mindsdb.interfaces.storage.db import session, Predictor

        create_process_mark('predict')
        original_name = name
        name = f'{company_id}@@@@@{name}'

        if name not in self.predictor_cache:
            # Clear the cache entirely if we have less than 1.2 GB left
            if psutil.virtual_memory().available < 1.2 * pow(10, 9):
                self.predictor_cache = {}

            predictor_record = Predictor.query.filter_by(company_id=company_id, name=original_name, is_custom=False).first()
            if predictor_record.data['status'] == 'complete':
                self.fs_store.get(name, f'predictor_{company_id}_{predictor_record.id}', self.config['paths']['predictors'])
                self.predictor_cache[name] = {
                    'predictor': mindsdb_native.Predictor(name=name, run_env={'trigger': 'mindsdb'}),
                    'created': datetime.datetime.now()
                }
                predictor = mindsdb_native.Predictor(name=name, run_env={'trigger': 'mindsdb'})

        if isinstance(when_data, dict) and 'kwargs' in when_data and 'args' in when_data:
            data_source = getattr(mindsdb_datasources, when_data['class'])(*when_data['args'], **when_data['kwargs'])
        else:
            # @TODO: Replace with Datasource
            try:
                data_source = pd.DataFrame(when_data)
            except Exception:
                data_source = when_data

        predictor = self.predictor_cache[name]['predictor']
        predictions = predictor.predict(
            when_data=data_source,
            **kwargs
        )
        del self.predictor_cache[name]
        if pred_format == 'explain' or pred_format == 'new_explain':
            predictions = [p.explain() for p in predictions]
        elif pred_format == 'dict':
            predictions = [p.as_dict() for p in predictions]
        elif pred_format == 'dict&explain':
            predictions = [[p.as_dict() for p in predictions], [p.explain() for p in predictions]]
        else:
            delete_process_mark('predict')
            raise Exception(f'Unkown predictions format: {pred_format}')

        delete_process_mark('predict')

        return predictions

    def analyse_dataset(self, ds, company_id=None):
        from mindsdb_datasources import FileDS, ClickhouseDS, MariaDS, MySqlDS, PostgresDS, MSSQLDS, MongoDS, SnowflakeDS, AthenaDS
        from mindsdb_native import F

        create_process_mark('analyse')
        ds = eval(ds['class'])(*ds['args'], **ds['kwargs'])
        analysis = F.analyse_dataset(ds)

        delete_process_mark('analyse')
        return analysis

    def get_model_data(self, name, db_fix=True, company_id=None):
        from mindsdb_native import F
        from mindsdb_native.libs.constants.mindsdb import DATA_SUBTYPES
        from mindsdb.interfaces.storage.db import session, Predictor, Datasource
        import torch
        import gc

        if '@@@@@' in name:
            name = name.split('@@@@@')[1]

        original_name = name
        name = f'{company_id}@@@@@{name}'

        predictor_record = Predictor.query.filter_by(company_id=company_id, name=original_name, is_custom=False).first()
        linked_data_source = Datasource.query.filter_by(company_id=company_id, id=predictor_record.datasource_id).first()
        predictor_record = self._try_outdate_db_status(predictor_record)
        model = predictor_record.data
        if model is None or model['status'] == 'training':
            try:
                self.fs_store.get(name, f'predictor_{company_id}_{predictor_record.id}', self.config['paths']['predictors'])
                new_model_data = F.get_model_data(name)
            except Exception:
                new_model_data = None

            try:
                torch.cuda.empty_cache()
            except Exception:
                pass
            gc.collect()

            if predictor_record.data is None or (new_model_data is not None and len(new_model_data) > len(predictor_record.data)):
                predictor_record.data = new_model_data
                model = new_model_data
                session.commit()

            if predictor_record.data is None:
                if new_model_data is None:
                    predictor_record.data = {"name": original_name, "status": "error"}
                    model = {"name": original_name, "status": "error"}
                elif len(new_model_data) > len(predictor_record.data):
                    predictor_record.data = new_model_data
                    model = new_model_data
                session.commit()

        # Make some corrections for databases not to break when dealing with empty columns
        if db_fix:
            data_analysis = model['data_analysis_v2']
            for column in model['columns']:
                analysis = data_analysis.get(column)
                if isinstance(analysis, dict) and (len(analysis) == 0 or analysis.get('empty', {}).get('is_empty', False)):
                    data_analysis[column]['typing'] = {
                        'data_subtype': DATA_SUBTYPES.INT
                    }

        model['created_at'] = str(parse_datetime(str(predictor_record.created_at).split('.')[0]))
        model['updated_at'] = str(parse_datetime(str(predictor_record.updated_at).split('.')[0]))
        model['predict'] = predictor_record.to_predict
        model['update'] = predictor_record.update_status
        model['name'] = predictor_record.name
        model['data_source_name'] = linked_data_source.name if linked_data_source else None
        return model

    def get_models(self, company_id=None):
        from mindsdb.interfaces.storage.db import session, Predictor

        models = []
        predictor_records = Predictor.query.filter_by(company_id=company_id, is_custom=False)
        predictor_names = [
            x.name for x in predictor_records
        ]
        for model_name in predictor_names:
            try:
                model_data = self.get_model_data(model_name, db_fix=False, company_id=company_id)

                reduced_model_data = {}

                for k in ['name', 'version', 'is_active', 'predict', 'status', 'current_phase', 'accuracy', 'data_source', 'update', 'data_source_name']:
                    reduced_model_data[k] = model_data.get(k, None)

                for k in ['train_end_at', 'updated_at', 'created_at']:
                    reduced_model_data[k] = model_data.get(k, None)
                    if reduced_model_data[k] is not None:
                        try:
                            reduced_model_data[k] = parse_datetime(str(reduced_model_data[k]).split('.')[0])
                        except Exception as e:
                            # @TODO Does this ever happen
                            log.error(f'Date parsing exception while parsing: {k} in get_models: ', e)
                            reduced_model_data[k] = parse_datetime(str(reduced_model_data[k]))

                models.append(reduced_model_data)
            except Exception as e:
                log.error(f"Can't list data for model: '{model_name}' when calling `get_models(), error: {e}`")
        return models

    def delete_model(self, name, company_id=None):
        from mindsdb_native import F
        from mindsdb_native.libs.constants.mindsdb import DATA_SUBTYPES
        from mindsdb.interfaces.storage.db import session, Predictor

        original_name = name
        name = f'{company_id}@@@@@{name}'

        predictor_record = Predictor.query.filter_by(company_id=company_id, name=original_name, is_custom=False).first()
        id = predictor_record.id
        session.delete(predictor_record)
        session.commit()
        F.delete_model(name)
        DatabaseWrapper(company_id).unregister_predictor(name)
        self.fs_store.delete(f'predictor_{company_id}_{id}')
        return 0

    def update_model(self, name, company_id=None):
        from mindsdb_native import F
        from mindsdb_worker.updater.update_model import update_model
        from mindsdb.interfaces.storage.db import session, Predictor
        from mindsdb.interfaces.datastore.datastore import DataStore, DataStoreWrapper
        from mindsdb_native import __version__ as native_version
        from mindsdb import __version__ as mindsdb_version

        original_name = name
        name = f'{company_id}@@@@@{name}'

        try:
            predictor_record = Predictor.query.filter_by(company_id=company_id, name=original_name, is_custom=False).first()

            predictor_record.update_status = 'updating'

            session.commit()

            update_model(name, original_name, self.delete_model, F.rename_model, self.learn_for_update, self._lock_context, company_id, self.config['paths']['predictors'], predictor_record, self.fs_store, DataStoreWrapper(DataStore(), company_id))

            predictor_record = Predictor.query.filter_by(company_id=company_id, name=original_name, is_custom=False).first()

            predictor_record.native_version = native_version
            predictor_record.mindsdb_version = mindsdb_version
            predictor_record.update_status = 'up_to_date'

            session.commit()
            
        except Exception as e:
            log.error(e)
            predictor_record.update_status = 'update_failed'
            session.commit()
            return str(e)
        
        return 'Updated successfully'


'''
Notes: Remove ray from actors are getting stuck
try:
    from mindsdb_worker.cluster.ray_controller import ray_ify
    import ray
    try:
        ray.init(ignore_reinit_error=True, address='auto')
    except Exception:
        ray.init(ignore_reinit_error=True)
    ModelController = ray_ify(ModelController)
except Exception as e:
    pass
'''



from mindsdb.api.mongo.classes import Responder
from mindsdb.interfaces.datastore.datastore import DataStoreWrapper
from mindsdb.interfaces.model.model_interface import ModelInterfaceWrapper


class Responce(Responder):
    def when(self, query):
        return 'company_id' in query

    def result(self, query, request_env, mindsdb_env, session):
        company_id = query.get('company_id')
        need_response = query.get('need_response', False)

        mindsdb_env['company_id'] = company_id
        mindsdb_env['data_store'] = DataStoreWrapper(
            data_store=mindsdb_env['data_store'],
            company_id=company_id
        )
        mindsdb_env['mindsdb_native'] = ModelInterfaceWrapper(
            model_interface=mindsdb_env['mindsdb_native'],
            company_id=company_id
        )

        if need_response:
            return {'ok': 1}

        return None


responder = Responce()



import re

import certifi
from pymongo import MongoClient

from mindsdb.integrations.base import Integration


class MongoConnectionChecker:
    def __init__(self, **kwargs):
        self.host = kwargs.get("host")
        self.port = kwargs.get("port", 27017)
        self.user = kwargs.get("username")
        self.password = kwargs.get("password")

    def _handle_params(self):
        kwargs = {}
        if isinstance(self.user, str) and len(self.user) > 0:
            kwargs['username'] = self.user

        if isinstance(self.password, str) and len(self.password) > 0:
            kwargs['password'] = self.password

        if re.match(r'\/\?.*tls=true', self.host.lower()):
            kwargs['tls'] = True

        if re.match(r'\/\?.*tls=false', self.host.lower()):
            kwargs['tls'] = False

        if re.match(r'.*\.mongodb.net', self.host.lower()) and kwargs.get('tls', None) is None:
            kwargs['tlsCAFile'] = certifi.where()
            if kwargs.get('tls', None) is None:
                kwargs['tls'] = True

        return kwargs

    def check_connection(self):
        try:
            advanced_conn_params = self._handle_params()
            server = MongoClient(self.host,
                                 port=self.port,
                                 serverSelectionTimeoutMS=5000,
                                 **advanced_conn_params
                                 )
            server.server_info()
            connected = True
        except Exception:
            connected = False
        return connected


class MongoDB(Integration, MongoConnectionChecker):
    def __init__(self, config, name, db_info):
        super().__init__(config, name)
        self.user = db_info.get('user', 'default')
        self.password = db_info.get('password', None)
        self.host = db_info.get('host')
        self.port = db_info.get('port', 27017)

    def _query(self, query):
        return None

    def setup(self):
        pass

    def register_predictors(self, model_data_arr):
        pass

    def unregister_predictor(self, name):
        pass



from mindsdb.api.mysql.mysql_proxy.datahub.information_schema import InformationSchema
from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.mindsdb_datanode import MindsDBDataNode
from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.datasource_datanode import DataSourceDataNode
from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.integration_datanode import IntegrationDataNode
from mindsdb.interfaces.database.integrations import get_db_integrations


def init_datahub(model_interface, custom_models, ai_table, data_store, company_id=None):
    datahub = InformationSchema()

    datahub.add({
        'mindsdb': MindsDBDataNode(model_interface, custom_models, ai_table, data_store, company_id),
        'datasource': DataSourceDataNode(data_store)
    })

    integrations = get_db_integrations(company_id).keys()
    for key in integrations:
        datahub.add({
            key: IntegrationDataNode(key, data_store)
        })

    return datahub



"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""

import re

from mindsdb_sql import parse_sql
from mindsdb_sql.planner import plan_query
from mindsdb_sql.parser.ast import Join, Identifier, Operation, Constant, UnaryOperation, BinaryOperation, OrderBy
from mindsdb_sql.planner.steps import (
    FetchDataframeStep,
    ApplyPredictorStep,
    ApplyPredictorRowStep,
    JoinStep,
    ProjectStep,
    FilterStep
)
from mindsdb_sql.parser.dialects.mindsdb.latest import Latest

from mindsdb.api.mysql.mysql_proxy.classes.com_operators_new import operator_map as new_operator_map
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import TYPES
from mindsdb.api.mysql.mysql_proxy.utilities import log
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import ERR
from mindsdb.interfaces.ai_table.ai_table import AITableStore

from mindsdb.api.mysql.mysql_proxy.utilities.sql import to_moz_sql_struct, plain_where_conditions


class SQLQuery():
    raw = ''
    struct = {}
    result = None

    def __init__(self, sql, integration=None, database=None, datahub=None):
        self.integration = integration
        if not database:
            self.database = 'mindsdb'
        else:
            self.database = database
        self.datahub = datahub

        self.ai_table = None

        # 'offset x, y' - specific just for mysql, parser dont understand it
        sql = re.sub(r'\n?limit([\n\d\s]*),([\n\d\s]*)', ' limit \g<1> offset \g<1> ', sql)

        self.raw = sql
        self._parse_query(sql)

    def fetch(self, datahub, view='list'):
        data = self.fetched_data

        if view == 'list':
            self.result = self._make_list_result_view(data)
        else:
            raise Exception('Only "list" view supported atm')

        return {
            'success': True,
            'result': self.result
        }

    def _parse_query(self, sql):
        def get_preditor_alias(step, mindsdb_database):
            return (mindsdb_database, step.predictor, step.alias)

        def get_table_alias(table_obj, default_db_name):
            # (database, table, alias)
            if len(table_obj.parts) > 2:
                raise Exception(f'Table name must contain no more than 2 parts. Got name: {table_obj.parts}')
            elif len(table_obj.parts) == 1:
                name = (default_db_name, table_obj.parts[0])
            else:
                name = tuple(table_obj.parts)
            name = name + (table_obj.alias,)
            return name

        mindsdb_sql_struct = parse_sql(sql, dialect='mindsdb')

        integrations_names = self.datahub.get_integrations_names()
        integrations_names.append('INFORMATION_SCHEMA')

        mindsdb_datanode = self.datahub.get(self.database)

        models = mindsdb_datanode.model_interface.get_models()
        predictor_metadata = {}
        for model in models:
            model_meta = mindsdb_datanode.model_interface.get_model_data(name=model['name'])
            window = model_meta.get('timeseries', {}).get('user_settings', {}).get('window')
            predictor_metadata[model_meta['name']] = {'timeseries': False}
            if window is not None:
                order_by = model_meta.get('timeseries', {}).get('user_settings', {}).get('order_by')[0]
                group_by = model_meta.get('timeseries', {}).get('user_settings', {}).get('group_by')[0]
                predictor_metadata[model_meta['name']] = {
                    'timeseries': True,
                    'window': window,
                    'order_by_column': order_by,
                    'group_by_column': group_by
                }

        plan = plan_query(
            mindsdb_sql_struct,
            integrations=integrations_names,
            predictor_namespace=self.database,
            predictor_metadata=predictor_metadata
        )
        steps_data = []

        for i, step in enumerate(plan.steps):
            data = []
            if isinstance(step, FetchDataframeStep):
                dn = self.datahub.get(step.integration)
                query = step.query

                data = dn.select_query(
                    query=query
                )
                table_alias = get_table_alias(step.query.from_table, self.database)
                data = [{table_alias: x} for x in data]
            elif isinstance(step, ApplyPredictorRowStep):
                dn = self.datahub.get(self.database)
                where_data = step.row_dict
                data = dn.select(
                    table=step.predictor,
                    columns=None,
                    where_data=where_data,
                    where={}
                )
                data = [{get_preditor_alias(step, self.database): x} for x in data]
            elif isinstance(step, ApplyPredictorStep):
                dn = self.datahub.get(self.database)
                where_data = []
                for row in steps_data[step.dataframe.step_num]:
                    new_row = {}
                    for table_name in row:
                        keys_intersection = set(new_row) & set(row[table_name])
                        if len(keys_intersection) > 0:
                            raise Exception(
                                f'The predictor got two identical keys from different datasources: {keys_intersection}'
                            )
                        new_row.update(row[table_name])
                    where_data.append(new_row)

                is_timeseries = predictor_metadata[step.predictor]['timeseries']
                if is_timeseries:
                    for row in where_data:
                        row['make_predictions'] = False

                data = dn.select(
                    table=step.predictor,
                    columns=None,
                    where_data=where_data,
                    where={},
                    is_timeseries=is_timeseries
                )
                data = [{get_preditor_alias(step, self.database): x} for x in data]
            elif isinstance(step, JoinStep):
                left_data = steps_data[step.left.step_num]
                right_data = steps_data[step.right.step_num]
                if step.query.condition is None:
                    # line-to-line join
                    if len(left_data) != len(right_data):
                        raise Exception('wrong data length')
                    data = []
                    left_alias = step.query.left.alias
                    right_alias = step.query.right.alias
                    for i in range(len(left_data)):
                        data.append({
                            left_alias: left_data[i][left_alias],
                            right_alias: right_data[i][right_alias]
                        })
                else:
                    raise Exception('Unknown join type')
            elif isinstance(step, FilterStep):
                raise Exception('FilterStep not implemented')
            elif isinstance(step, ProjectStep):
                step_data = steps_data[step.dataframe.step_num]
                row = step_data[0]  # TODO if rowcount = 0
                tables_columns = {}
                for table_name in row:
                    tables_columns[table_name] = list(row[table_name].keys())

                columns_list = []
                for column_full_name in step.columns:
                    table_name = None
                    if column_full_name != '*':
                        column_name_parts = column_full_name.split('.')
                        if len(column_name_parts) > 2:
                            raise Exception(f'Column name must contain no more than 2 parts. Got name: {column_full_name}')
                        elif len(column_name_parts) == 1:
                            column_name = column_name_parts[0]
                            column_alias = step.aliases.get(column_full_name)

                            appropriate_table = None
                            for table_name, table_columns in tables_columns.items():
                                if column_name in table_columns:
                                    if appropriate_table is not None:
                                        raise Exception('Fount multiple appropriate tables for column {column_name}')
                                    else:
                                        appropriate_table = table_name
                            if appropriate_table is None:
                                raise Exception(f'Can not find approproate table for column {column_name}')
                            columns_list.append(appropriate_table + (column_name, column_alias))
                        elif len(column_name_parts) == 2:
                            table_name_or_alias = column_name_parts[0]
                            column_name = column_name_parts[1]
                            column_alias = step.aliases.get(column_full_name)

                            appropriate_table = None
                            for table_name, table_columns in tables_columns.items():
                                checkig_table_name_or_alias = table_name[2] or table_name[1]
                                if table_name_or_alias == checkig_table_name_or_alias:
                                    if column_name not in table_columns:
                                        raise Exception(f'Can not find column "{column_name}" in table "{table_name}"')
                                    appropriate_table = table_name
                                    break
                            if appropriate_table is None:
                                raise Exception(f'Can not find approproate table for column {column_name}')
                            columns_list.append(appropriate_table + (column_name, column_alias))
                        else:
                            raise Exception('Undefined column name')
                    else:
                        for table_name, table_columns in tables_columns.items():
                            for column_name in table_columns:
                                columns_list.append(table_name + (column_name, None))

                self.columns_list = columns_list
                data = step_data
            steps_data.append(data)

        self.fetched_data = steps_data[-1]

    def _apply_where_filter(self, row, where):
        if isinstance(where, Identifier):
            return row[where.value]
        elif isinstance(where, Constant):
            return where.value
        elif not isinstance(where, (UnaryOperation, BinaryOperation)):
            Exception(f'Unknown operation type: {where}')

        op_fn = new_operator_map.get(where.op)
        if op_fn is None:
            raise Exception(f'unknown operator {where.op}')

        args = [self._apply_where_filter(row, arg) for arg in where.args]
        result = op_fn(*args)
        return result

    def _make_list_result_view(self, data):
        result = []
        for row in data:
            data_row = []
            for column_record in self.columns_list:
                table_name = column_record[:3]
                column_name = column_record[3]
                data_row.append(row[table_name][column_name])
            result.append(data_row)
        return result

    @property
    def columns(self):
        result = []
        for column_record in self.columns_list:
            result.append({
                'database': column_record[0] or self.database,
                #  TODO add 'original_table'
                'table_name': column_record[1],
                'name': column_record[3],
                'alias': column_record[4] or column_record[3],
                # NOTE all work with text-type, but if/when wanted change types to real,
                # it will need to check all types casts in BinaryResultsetRowPacket
                'type': TYPES.MYSQL_TYPE_VAR_STRING
            })
        return result
